/* -*- c++ -*- */
/*
 * Expresso, a C++ Array template class library
 * Copyright (C) 1998,2004 Oskar Enoksson (enok@lysator.liu.se)
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
#ifndef EXPRESSO_ITERATOR_HDR
#define EXPRESSO_ITERATOR_HDR

#include "promote.h"

namespace esso {

  template<class Type>
  struct Iterator;

  // The general iterator type
  template<int N, class Type>
  class Array<N,Iterator<Type> > {
  public:
    typedef typename Array<N,Type>::T T;
    typedef Array<N,Iterator<Type> > IteratorType;
    enum { hasdata=0 };

    inline IXTYPE L(DIMT i) const { return src.L(i); }
    inline IXTYPE U(DIMT i) const { return src.U(i); }
    inline IXTYPE Len(DIMT i) const { return U(i)-L(i); }
    inline void Step(DIMT n) { origo+=src.Stride(n); }
    inline void Step(DIMT n, IXTYPE i) { origo+=src.Stride(n)*i; }
    inline T CGet() const { return *origo; }
    inline T CGet(IXTYPE i) const { return origo[i*src.Stride(0)]; }
    inline T & Get() const { return *origo; }
    inline T & Get(IXTYPE i) const { return origo[i*src.Stride(0)]; }

    inline Array(const Array<N,Type> &src0): src(src0), origo(src0.Origo()) {}
  private:
    const Array<N,Type> & src;
    T * origo;
  };

  template<class T00>
  struct Fix;

  // Specialization of iterator for Fix arrays
  template<int N, class T0>
  class Array<N,Iterator<Fix<T0> > > {
    typedef Fix<T0> Type;
  public:
    typedef typename Array<N,Type>::T T;
    typedef Array<N,Iterator<Type> > IteratorType;
    enum { hasdata=0 };

    inline IXTYPE L(int i) const { return src.L(i); }
    inline IXTYPE U(int i) const { return src.U(i); }
    inline IXTYPE Len(int i) const { return U(i)-L(i); }
    inline void Step(int n) { origo+=src.Stride(n); }
    inline void Step(int n, IXTYPE i) { origo+=src.Stride(n)*i; }
    inline T CGet() const { return *origo; }
    inline T CGet(IXTYPE i) const { return origo[i]; }
    inline T & Get() const { return *origo; }
    inline T & Get(IXTYPE i) const { return origo[i]; }

    inline Array(const Array<N,Type> &src0): src(src0), origo(src0.Origo()) {}
  private:
    const Array<N,Type> & src;
    T * origo;
  };

  template<class T00>
  class Scalar;

  template<int N, class T0>
  class Array<N,Scalar<T0> > {
  public:
    typedef typename TmplTools::ArrayInfo<T0>::T T;
    typedef Array<N,Scalar<T0> > IteratorType;
    enum { hasdata=0 };

    inline IXTYPE L(DIMT i) const { return IXTYPE_MIN; }
    inline IXTYPE U(DIMT i) const { return IXTYPE_MAX; }
    inline IXTYPE Len(DIMT i) const { return U(i)-L(i); }
    inline void Step(DIMT n) {}
    inline void Step(DIMT n, IXTYPE i) {}
    inline T CGet() const { return src; }
    inline T CGet(IXTYPE i) const { return src; }

    inline Array(const T &src0): src(src0) {}
  private:
    const T src;
  };

  // Unary operator Type definition
  template<class Op, class TypeS>
  class Uop;

  template<int N, class Op, class TypeS>
  class Array<N,Uop<Op,TypeS> > {
  public:
    typedef typename UPromote<Op, typename Array<N,TypeS>::T>::T T;
    typedef Array<N,Uop<Op,TypeS> > IteratorType;
    enum { hasdata=0 };

    inline IXTYPE L(DIMT i) const { return src.L(i); }
    inline IXTYPE U(DIMT i) const { return src.U(i); }
    inline IXTYPE Len(DIMT i) const { return U(i)-L(i); }
    inline void Step(DIMT n) { src.Step(n); }
    inline void Step(DIMT n, IXTYPE i) { src.Step(n,i); }
    inline T CGet() const { 
      return UPromote<Op,typename Array<N,TypeS>::T>::Do(src.CGet()); }
    inline T CGet(IXTYPE i) const { 
      return UPromote<Op,typename Array<N,TypeS>::T>::Do(src.CGet(i)); }
      
    inline Array(const Array<N,TypeS> &src0): src(src0) {}
  private:
    typename Array<N,TypeS>::IteratorType src;
  };

  // Binary operator Type definition
  template<class Op, class TypeL, class TypeR>
  class Bop;

  template<int N, class Op, class TypeL, class TypeR>
  class Array<N,Bop<Op,TypeL,TypeR> > {
  public:
    typedef typename BPromote<Op,typename Array<N,TypeL>::T,
			      typename Array<N,TypeR>::T>::T T;
    typedef Array<N,Bop<Op,TypeL,TypeR> > IteratorType;
    enum { hasdata=0 };

    inline IXTYPE L(DIMT i) const { return std::max(srcl.L(i),srcr.L(i)); }
    inline IXTYPE U(DIMT i) const { return std::min(srcl.U(i),srcr.U(i)); }
    inline IXTYPE Len(DIMT i) const { return U(i)-L(i); }
    inline void Step(DIMT n) { srcl.Step(n); srcr.Step(n); }
    inline void Step(DIMT n, IXTYPE i) { srcl.Step(n,i); srcr.Step(n,i); }
    inline T CGet() const { 
      return BCalculate<Op,
	typename Array<N,TypeL>::T,
	typename Array<N,TypeR>::T>::Do(srcl.CGet(),srcr.CGet()); 
    }
    inline T CGet(IXTYPE i) const {
      return BCalculate<Op,
	typename Array<N,TypeL>::T,
	typename Array<N,TypeR>::T>::Do(srcl.CGet(i),srcr.CGet(i));
    }

    inline Array(const Array<N,TypeL> &srcl0, const Array<N,TypeR> &srcr0):
      srcl(srcl0), srcr(srcr0) {}
  private:
    typename Array<N,TypeL>::IteratorType srcl;
    typename Array<N,TypeR>::IteratorType srcr;
  };

  // Reducing operator Type definition  
  template<class Op, class TypeS>
  class Rop;

  template<int N, class Op, class TypeS>
  class Array<N,Rop<Op, TypeS> > {
  public:
    typedef typename RPromote<Op, typename Array<N+1,TypeS>::T>::T T;
    typedef Array<N,Rop<Op,TypeS> > IteratorType;
    enum { hasdata=0 };

    inline IXTYPE L(DIMT i) const { return src.L(i+1); }
    inline IXTYPE U(DIMT i) const { return src.U(i+1); }
    inline IXTYPE Len(DIMT i) const { return U(i+1)-L(i+1); }
    inline void Step(DIMT n) { src.Step(n+1); }
    inline void Step(DIMT n, IXTYPE i) { src.Step(n+1,i); }

    inline T CGet() const {
      return RPromote<Op,typename Array<N+1,TypeS>::T>::Do(src); }
    inline T CGet(IXTYPE i) const {
      src.Step(1,i);
      T ret=RPromote<Op,typename Array<N+1,TypeS>::T>::Do(src);
      src.Step(1,-i);
      return ret;
    }

    inline Array(const Array<N+1,TypeS> &src0): src(src0) {}
  private:
    mutable typename Array<N+1,TypeS>::IteratorType src;
  };
}

#endif
