/* -*- c++ -*- */
/*
 * Expresso, a C++ Array template class library
 * Copyright (C) 1998,2004 Oskar Enoksson (enok@lysator.liu.se)
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
#ifndef EXPRESSO_STANDARDOP_HDR
#define EXPRESSO_STANDARDOP_HDR

#include "iterator.h"

namespace esso {
  
  ///@name Arithmetic Array operations
  //\anchor arithmetic_array_operations
  //@{

  struct cAdd;
  template<class TL, class TR>
  struct BCalculate<cAdd,TL,TR> {
    static inline typename BPromote<cAdd,TL,TR>::T
    Do(const TL l, const TR r) { return l+r; }
  };
  /**@arg \a l Left Array operand
   *@arg \a r Right Array operand
   *@return Array where each element refers to the sum of the 
   * corresponding elements in \a l and \a r */
  template<int N, class TypeL, class TypeR>
  inline Array<N,Bop<cAdd,TypeL,TypeR> >
  operator +
  (const Array<N,TypeL> &l,
   const Array<N,TypeR> &r) {
    return Array<N,Bop<cAdd,TypeL,TypeR> >(l,r);
  }
  /**@arg \a l Left Array operand
   *@arg \a r Right scalar operand
   *@return Array where each element refers to the sum of the 
   * corresponding elements in \a l and \a r */
  template<int N, class TypeL, class TR>
  inline Array<N,Bop<cAdd,TypeL,Scalar<TR> > >
  operator +
  (const Array<N,TypeL> &l,
   const TR &r) {
    return Array<N,Bop<cAdd,TypeL,Scalar<TR> > >(l,Array<N,Scalar<TR> >(r));
  }
  /**@arg \a l Left scalar operand
   *@arg \a r Right Array operand
   *@return Array where each element refers to the sum of the 
   * corresponding elements in \a l and \a r */
  template<int N, class TL, class TypeR>
  inline Array<N,Bop<cAdd,Scalar<TL>,TypeR > >
  operator +
  (const TL &l,
   const Array<N,TypeR> &r) {
    return Array<N,Bop<cAdd,Scalar<TL>,TypeR > >(Array<N,Scalar<TL> >(l),r);
  }

  struct cSub;
  template<class TL, class TR>
  struct BCalculate<cSub,TL,TR> {
    static inline typename BPromote<cSub,TL,TR>::T
    Do(const TL l, const TR r) { return l-r; }
  };
  /**@arg \a l Left Array operand
   *@arg \a r Right Array operand
   *@return Array where each element refers to the difference of the 
   * corresponding elements in \a l and \a r */
  template<int N, class TypeL, class TypeR>
  inline Array<N,Bop<cSub,TypeL,TypeR> >
  operator -
  (const Array<N,TypeL> &l,
   const Array<N,TypeR> &r) {
    return Array<N,Bop<cSub,TypeL,TypeR> >(l,r);
  }
  /**@arg \a l Left Array operand
   *@arg \a r Right scalar operand
   *@return Array where each element refers to the difference of the 
   * corresponding elements in \a l and \a r */
  template<int N, class TypeL, class TR>
  inline Array<N,Bop<cSub,TypeL,Scalar<TR> > >
  operator -
  (const Array<N,TypeL> &l,
   const TR &r) {
    return Array<N,Bop<cSub,TypeL,Scalar<TR> > >(l,Array<N,Scalar<TR> >(r));
  }
  /**@arg \a l Left scalar operand
   *@arg \a r Right Array operand
   *@return Array where each element refers to the difference of the 
   * corresponding elements in \a l and \a r */
  template<int N, class TL, class TypeR>
  inline Array<N,Bop<cSub,Scalar<TL>,TypeR > >
  operator -
  (const TL &l,
   const Array<N,TypeR> &r) {
    return Array<N,Bop<cSub,Scalar<TL>,TypeR > >(Array<N,Scalar<TL> >(l),r);
  }

  struct cMul;
  template<class TL, class TR>
  struct BCalculate<cMul,TL,TR> {
    static inline typename BPromote<cMul,TL,TR>::T
    Do(const TL l, const TR r) { return l*r; }
  };
  /**@arg \a l Left Array operand
   *@arg \a r Right Array operand
   *@return Array where each element refers to the product of the 
   * corresponding elements in \a l and \a r */
  template<int N, class TypeL, class TypeR>
  inline Array<N,Bop<cMul,TypeL,TypeR> >
  operator *
  (const Array<N,TypeL> &l,
   const Array<N,TypeR> &r) {
    return Array<N,Bop<cMul,TypeL,TypeR> >(l,r);
  }
  /**@arg \a l Left Array operand
   *@arg \a r Right scalar operand
   *@return Array where each element refers to the product of the 
   * corresponding elements in \a l and \a r */
  template<int N, class TypeL, class TR>
  inline Array<N,Bop<cMul,TypeL,Scalar<TR> > >
  operator *
  (const Array<N,TypeL> &l,
   const TR &r) {
    return Array<N,Bop<cMul,TypeL,Scalar<TR> > >(l,Array<N,Scalar<TR> >(r));
  }
  /**@arg \a l Left scalar operand
   *@arg \a r Right Array operand
   *@return Array where each element refers to the product of the 
   * corresponding elements in \a l and \a r */
  template<int N, class TL, class TypeR>
  inline Array<N,Bop<cMul,Scalar<TL>,TypeR > >
  operator *
  (const TL &l,
   const Array<N,TypeR> &r) {
    return Array<N,Bop<cMul,Scalar<TL>,TypeR > >(Array<N,Scalar<TL> >(l),r);
  }

  struct cDiv;
  template<class TL, class TR>
  struct BCalculate<cDiv,TL,TR> {
    static inline typename BPromote<cDiv,TL,TR>::T
    Do(const TL l, const TR r) { return l/r; }
  };
  /**@arg \a l Left Array operand
   *@arg \a r Right Array operand
   *@return Array where each element refers to the quotient of the 
   * corresponding elements in \a l and \a r */
  template<int N, class TypeL, class TypeR>
  inline Array<N,Bop<cDiv,TypeL,TypeR> >
  operator /
  (const Array<N,TypeL> &l,
   const Array<N,TypeR> &r) {
    return Array<N,Bop<cDiv,TypeL,TypeR> >(l,r);
  }
  /**@arg \a l Left Array operand
   *@arg \a r Right scalar operand
   *@return Array where each element refers to the quotient of the 
   * corresponding elements in \a l and \a r */
  template<int N, class TypeL, class TR>
  inline Array<N,Bop<cDiv,TypeL,Scalar<TR> > >
  operator /
  (const Array<N,TypeL> &l,
   const TR &r) {
    return Array<N,Bop<cDiv,TypeL,Scalar<TR> > >(l,Array<N,Scalar<TR> >(r));
  }
  /**@arg \a l Left scalar operand
   *@arg \a r Right Array operand
   *@return Array where each element refers to the quotient of the 
   * corresponding elements in \a l and \a r */
  template<int N, class TL, class TypeR>
  inline Array<N,Bop<cDiv,Scalar<TL>,TypeR > >
  operator /
  (const TL &l,
   const Array<N,TypeR> &r) {
    return Array<N,Bop<cDiv,Scalar<TL>,TypeR > >(Array<N,Scalar<TL> >(l),r);
  }

  struct cMax;
  template<class TL, class TR>
  struct BCalculate<cMax,TL,TR> {
    static inline typename BPromote<cMax,TL,TR>::T
    Do(const TL l, const TR r) { return
      l>r?l:r;
    }
  };
  /**@arg \a l Left Array operand
   *@arg \a r Right Array operand
   *@return Array where each element refers to the maximum of the 
   * corresponding elements in \a l and \a r */
  template<int N, class TypeL, class TypeR>
  inline Array<N,Bop<cMax,TypeL,TypeR> >
  Max
  (const Array<N,TypeL> &l,
   const Array<N,TypeR> &r) {
    return Array<N,Bop<cMax,TypeL,TypeR> >(l,r);
  }
  /**@arg \a l Left Array operand
   *@arg \a r Right scalar operand
   *@return Array where each element refers to the maximum of the 
   * corresponding elements in \a l and \a r */
  template<int N, class TypeL, class TR>
  inline Array<N,Bop<cMax,TypeL,Scalar<TR> > >
  Max
  (const Array<N,TypeL> &l,
   const TR &r) {
    return Array<N,Bop<cMax,TypeL,Scalar<TR> > >(l,Array<N,Scalar<TR> >(r));
  }
  /**@arg \a l Left scalar operand
   *@arg \a r Right Array operand
   *@return Array where each element refers to the maximum of the 
   * corresponding elements in \a l and \a r */
  template<int N, class TL, class TypeR>
  inline Array<N,Bop<cMax,Scalar<TL>,TypeR > >
  Max
  (const TL &l,
   const Array<N,TypeR> &r) {
    return Array<N,Bop<cMax,Scalar<TL>,TypeR > >(Array<N,Scalar<TL> >(l),r);
  }

  struct cMin;
  template<class TL, class TR>
  struct BCalculate<cMin,TL,TR> {
    static inline typename BPromote<cMin,TL,TR>::T
    Do(const TL l, const TR r) { return 
      l<r?l:r;
    }
  };
  /**@arg \a l Left Array operand
   *@arg \a r Right Array operand
   *@return Array where each element refers to the minimum of the 
   * corresponding elements in \a l and \a r */
  template<int N, class TypeL, class TypeR>
  inline Array<N,Bop<cMin,TypeL,TypeR> >
  Min
  (const Array<N,TypeL> &l,
   const Array<N,TypeR> &r) {
    return Array<N,Bop<cMin,TypeL,TypeR> >(l,r);
  }
  /**@arg \a l Left Array operand
   *@arg \a r Right scalar operand
   *@return Array where each element refers to the minimum of the 
   * corresponding elements in \a l and \a r */
  template<int N, class TypeL, class TR>
  inline Array<N,Bop<cMin,TypeL,Scalar<TR> > >
  Min
  (const Array<N,TypeL> &l,
   const TR &r) {
    return Array<N,Bop<cMin,TypeL,Scalar<TR> > >(l,Array<N,Scalar<TR> >(r));
  }
  /**@arg \a l Left scalar operand
   *@arg \a r Right Array operand
   *@return Array where each element refers to the minimum of the 
   * corresponding elements in \a l and \a r */
  template<int N, class TL, class TypeR>
  inline Array<N,Bop<cMin,Scalar<TL>,TypeR > >
  Min
  (const TL &l,
   const Array<N,TypeR> &r) {
    return Array<N,Bop<cMin,Scalar<TL>,TypeR > >(Array<N,Scalar<TL> >(l),r);
  }

  struct cNeg;
  template<class T0>
  struct UPromote<cNeg,T0> {
    typedef T0 T;
    static inline T Do(const T s) { return -s; }
  };
  /** Unary negation
   *@arg \a s Source array object
   *@return Array with each element from \a s negated */
  template<int N, class TypeS>
  inline Array<N,Uop<cNeg,TypeS> > 
  operator -
  (const Array<N,TypeS> &s) {
    return Array<N,Uop<cNeg,TypeS> >(s);
  }

  struct cSqrt;
  template<class T0>
  struct UPromote<cSqrt,T0> {
    typedef T0 T;
    static inline T Do(const T s) { return sqrt(s); }
  };
  /** Square root
   *@arg \a s Source array object
   *@return Array with each element square root of that from \a s */
  template<int N, class TypeS>
  inline Array<N,Uop<cSqrt,TypeS> > 
  Sqrt
  (const Array<N,TypeS> &s) {
    return Array<N,Uop<cSqrt,TypeS> >(s);
  }

  struct cSqr;
  template<class T0>
  struct UPromote<cSqr,T0> {
    typedef T0 T;
    static inline T Do(const T s) { return s*s; }
  };
  /** Square (power of two)
   *@arg \a s Source array object
   *@return Array with each element equal to the square of that from \a s */
  template<int N, class TypeS>
  inline Array<N,Uop<cSqr,TypeS> > 
  Sqr
  (const Array<N,TypeS> &s) {
    return Array<N,Uop<cSqr,TypeS> >(s);
  }

  struct cAbs;
  template<class T0>
  struct UPromote<cAbs,T0> {
    typedef T0 T;
    static inline T Do(const T s) { return s>0 ? s : -s; }
  };
  /** Absolute value
   *@arg \a s Source array object
   *@return Array with each element equal to the absolute value of that 
   * from \a s */
  template<int N, class TypeS>
  inline Array<N,Uop<cAbs,TypeS> > 
  Abs
  (const Array<N,TypeS> &s) {
    return Array<N,Uop<cAbs,TypeS> >(s);
  }

  template<class Op, class T0>
  struct RPromote { public: typedef T0 T; };

  struct cSum;
  template<class T0>
  struct RPromote<cSum,T0> {
    typedef T0 T;
    template<int N, class TypeS>
    static inline T Do(const Array<N,TypeS> &src) {
      IXTYPE u=src.U(0);
      IXTYPE l=src.L(0);
      T ret=T(0);
      for(IXTYPE i=l; i<u; i++)
	ret=ret+src.CGet(i);
      return ret; 
    }
  };
  /**@arg \a src Source array
   *@return Array where each element is the sum of all elements in
   * \a src along dimension 0 */
  template<int N, class TypeS>
  inline Array<N-1,Rop<cSum,TypeS> >
  Sum
  (const Array<N,TypeS> &s) {
    return Array<N-1,Rop<cSum,TypeS> >(s);
  }
  /**@arg \a src Source array 
   *@return Sum of all elements in \a src */
  template<class TypeS>
  inline typename RPromote<cSum,typename Array<1,TypeS>::T>::T
  Sum
  (const Array<1,TypeS> &src0) {
    typename Array<1,TypeS>::IteratorType src(src0);
    return RPromote<cSum,typename Array<1,TypeS>::T>::Do(src);
  }

  /**@arg \a src Source array
   *@return Sum of all elements in array */
  template<int N, class TypeS>
  inline typename Array<N,TypeS>::T Sumall(const Array<N,TypeS> &src) {
    return Sumall(Sum(src));  
  }
  template<class TypeS>
  inline typename Array<1,TypeS>::T Sumall(const Array<1,TypeS> &src) {
    return Sum(src); 
  }

  template<class T0>
  struct RPromote<cMax,T0> {
    typedef T0 T;
    template<int N, class TypeS>
    static inline T Do(const Array<N,TypeS> &src) {
      IXTYPE u=src.U(0);
      IXTYPE l=src.L(0);
      T ret=T(0);
      if(u>l)
	ret=src.CGet(l);
      for(IXTYPE i=l; i<u; i++)
	if(src.CGet(i)>ret)
	  ret=src.CGet(i);
      return ret; 
    }
  };
  /**@arg \a src Source array
   *@return Array where each element is the maximum of all elements in
   * \a src along dimension 0 */
  template<int N, class TypeS>
  inline Array<N-1,Rop<cMax,TypeS> >
  Max
  (const Array<N,TypeS> &s) {
    return Array<N-1,Rop<cMax,TypeS> >(s);
  }
  /**@arg \a src Source array 
   *@return Maximum of all elements in \a src */
  template<class TypeS>
  inline typename RPromote<cMax,typename Array<1,TypeS>::T>::T
  Max
  (const Array<1,TypeS> &src0) {
    typename Array<1,TypeS>::IteratorType src(src0);
    return RPromote<cMax,typename Array<1,TypeS>::T>::Do(src);
  }
  /**@arg \a src Source array
   *@return Maximum value of all elements in array */
  template<int N, class TypeS>
  inline typename Array<N,TypeS>::T Maxall(const Array<N,TypeS> &src) {
    return Maxall(Max(src));
  }
  template<class TypeS>
  inline typename Array<1,TypeS>::T Maxall(const Array<1,TypeS> &src) {
    return Max(src);
  }

  template<class T0>
  struct RPromote<cMin,T0> {
    typedef T0 T;
    template<int N, class TypeS>
    static inline T Do(const Array<N,TypeS> &src) {
      IXTYPE u=src.U(0);
      IXTYPE l=src.L(0);
      T ret=T(0);
      if(u>l)
	ret=src.CGet(l);
      for(IXTYPE i=l; i<u; i++)
	if(src.CGet(i)<ret)
	  ret=src.CGet(i);
      return ret; 
    }
  };
  /**@arg \a src Source array
   *@return Array where each element is the minimum of all elements in
   * \a src along dimension 0 */
  template<int N, class TypeS>
  inline Array<N-1,Rop<cMin,TypeS> >
  Min
  (const Array<N,TypeS> &s) {
    return Array<N-1,Rop<cMin,TypeS> >(s);
  }
  /**@arg \a src Source array 
   *@return Minimum of all elements in \a src */
  template<class TypeS>
  inline typename RPromote<cMin,typename Array<1,TypeS>::T>::T
  Min
  (const Array<1,TypeS> &src0) {
    typename Array<1,TypeS>::IteratorType src(src0);
    return RPromote<cMin,typename Array<1,TypeS>::T>::Do(src);
  }
  /**@arg \a src Source array
   *@return Minimum value of all elements in array */
  template<int N, class TypeS>
  inline typename Array<N,TypeS>::T Minall(const Array<N,TypeS> &src) {
    return Minall(Min(src));
  }
  template<class TypeS>
  inline typename Array<1,TypeS>::T Minall(const Array<1,TypeS> &src) {
    return Min(src);
  }
  //@}

  ///@name Logical Array operations
  //\anchor logical_array_operations
  //@{
  /**@arg \a src Source array
   *@arg \a n dimension to shift
   *@arg \a i number of index positions to shift
   *@return new shifted array refering to the same data as \a src
   *@see Array::Shift */
  template<int N, class Type>
  inline Array<N,Type>
  Shift(const Array<N,Type> &src, DIMT n, IXTYPE i) {
    return Array<N,Type>(src).Shift(n,i);
  }

  /**@arg \a src Source array
   *@arg \a ix vector of index positions to shift for each dimension
   *@return new shifted array refering to the same data as \a src
   *@see Array::Shift */
  template<int N, class Type>
  inline Array<N,Type>
  Shift(const Array<N,Type> &src, const Ix<N> &ix) {
    return Array<N,Type>(src).Shift(ix);
  }

  /**@arg \a src Source array
   *@arg \a n dimension to reverse
   *@return new reversed array refering to the same data as \a src
   *@see Array::Reverse */
  template<int N, class Type>
  inline Array<N,Type>
  Reverse(const Array<N,Type> &src, DIMT n) {
    return Array<N,Type>(src).Reverse(n);
  }

  /**@arg \a src Source array
   *@arg \a rx boolean vector indicating which dimensions to reverse
   *@return new reversed array refering to the same data as \a src
   *@see Array::Reverse */
  template<int N, class Type>
  inline Array<N,Type>
  Reverse(const Array<N,Type> &src, const Ix<N,bool> rx) {
    return Array<N,Type>(src).Reverse(rx);
  }

  /**@arg \a src Source array
   *@arg \a n dimension to flip
   *@return new flipped array refering to the same data as \a src
   *@see Array::Flip */
  template<int N, class Type>
  inline Array<N,Type>
  Flip(const Array<N,Type> &src, DIMT n) {
    return Array<N,Type>(src).Flip(n);
  }

  /**@arg \a src Source array
   *@arg \a rx boolean vector indicating which dimensions to flip
   *@return new flipped array refering to the same data as \a src
   *@see Array::Flip */
  template<int N, class Type>
  inline Array<N,Type>
  Flip(const Array<N,Type> &src, const Ix<N,bool> &rx) {
    return Array<N,Type>(src).Flip(rx);
  }

  /**@arg \a src Source array
   *@arg \a n dimension to restride
   *@arg \a s restride multiplier
   *@return new restrided array refering to the same data as \a src
   *@see Array::Restride */
  template<int N, class Type>
  inline Array<N,Type>
  Restride(const Array<N,Type> &src, DIMT n, IXTYPE s) {
    return Array<N,Type>(src).Restride(n,s);
  }

  /**@arg \a src Source array
   *@arg \a sx vector of restride multipliers for each dimension
   *@return new restrided array refering to the same data as \a src
   *@see Array::Restride */
  template<int N, class Type>
  inline Array<N,Type>
  Restride(const Array<N,Type> &src, const Ix<N> &sx) {
    return Array<N,Type>(src).Restride(sx);
  }

  /**@arg \a src Source array
   *@arg \a n dimension to restrict
   *@arg \a l0 new lower index bound
   *@return new restricted array refering to the same data as \a src
   *@see Array::RestrictL */
  template<int N, class Type>
  inline Array<N,Type>
  RestrictL(const Array<N,Type> &src, DIMT n, IXTYPE l0) {
    return Array<N,Type>(src).RestrictL(n,l0);
  }

  /**@arg \a src Source array
   *@arg \a n dimension to restrict
   *@arg \a u0 new upper index bound
   *@return new restricted array refering to the same data as \a src
   *@see Array::RestrictU */
  template<int N, class Type>
  inline Array<N,Type>
  RestrictU(const Array<N,Type> &src, DIMT n, IXTYPE u0) {
    return Array<N,Type>(src).RestrictU(n,u0);
  }

  /**@arg \a src Source array
   *@arg \a n dimension to restrict
   *@arg \a l0 new lower index bound
   *@arg \a u0 new upper index bound
   *@return new restricted array refering to the same data as \a src
   *@see Array::Restrict */
  template<int N, class Type>
  inline Array<N,Type>
  Restrict(const Array<N,Type> &src, DIMT n, IXTYPE l0, IXTYPE u0) {
    return Array<N,Type>(src).Restrict(n,l0,u0);
  }

  /**@arg \a src Source array
   *@arg \a l0 vector of new lower index bounds for each dimension
   *@return new restricted array refering to the same data as \a src
   *@see Array::RestrictL */
  template<int N, class Type>
  inline Array<N,Type>
  RestrictL(const Array<N,Type> &src, const Ix<N> &l0) {
    return Array<N,Type>(src).RestrictL(l0);
  }

  /**@arg \a src Source array
   *@arg \a u0 vector of new upper index bounds for each dimension
   *@return new restricted array refering to the same data as \a src
   *@see Array::RestrictU */
  template<int N, class Type>
  inline Array<N,Type>
  RestrictU(const Array<N,Type> &src, const Ix<N> &u0) {
    return Array<N,Type>(src).RestrictU(u0);
  }

  /**@arg \a src Source array
   *@arg \a l0 vector of new lower index bounds for each dimension
   *@arg \a u0 vector of new upper index bounds for each dimension
   *@return new restricted array refering to the same data as \a src
   *@see Array::Restrict */
  template<int N, class Type>
  inline Array<N,Type>
  Restrict(const Array<N,Type> &src, const Ix<N> &l0, const Ix<N> &u0) {
    return Array<N,Type>(src).Restrict(l0,u0);
  }

  /**@arg \a src Source array
   *@arg \a n1 \a n2 Dimensions to swap
   *@return new permuted array refering to the same data as \a src
   *@see Array::Permute */
  template<int N, class Type>
  inline Array<N,Type>
  Permute(const Array<N,Type> &src, DIMT n1, DIMT n2) {
    return Array<N,Type>(src).Permute(n1,n2);
  }

  /**@arg \a src Source array
   *@arg \a p permutation to apply to array dimensions
   *@return new permuted array refering to the same data as \a src
   *@see Array::Permute */
  template<int N, class Type>
  inline Array<N,Type>
  Permute(const Array<N,Type> &src, const Permutation<N> &p) {
    return Array<N,Type>(src).Permute(p);
  }

  /**@arg \a src Source array
   *@arg \a n Dimension to remove (how to cut a slice)
   *@arg \a i Index position along \a n (where to cut the slice)
   *@return new array refering to a slice of the same data as \a src
   *@see Array::Rmdim */
  template<int N, class Type>
  inline Array<N-1,Type>
  Rmdim(const Array<N,Type> &src, DIMT n, IXTYPE i) {
    return Array<N-1,Type>(src,n,i);
  }

  /**@arg \a src Source array
   *@arg \a l0 Lower index bound for new dimension
   *@arg \a u0 Upper index bound for new dimension
   *@return new array refering to the same data as \a src with a new
   * dummy dimension inserted at dimension 0
   *@see Array::Permute */
  template<int N, class Type>
  inline Array<N+1,Type>
  Extend(const Array<N,Type> &src, IXTYPE l0, IXTYPE u0) {
    return Array<N+1,Type>().Extend(src,l0,u0);
  }

  /**@arg \a l first array to compare
   *@arg \a r second array to compare
   *@return true iff the index bounds are identical for \a l and \a r */
  template<int N, class TypeL, class TypeR>
  inline bool Matches(const Array<N,TypeL> &l, const Array<N,TypeR> &r) {
    for(int n=0; n<N; n++)
      if(l.L(n)!=r.L(n) || l.U(n)!=r.U(n))
	return false;
    return true;
  }
  //@}
}

#endif
