/* -*- c++ -*- */
/*
 * Expresso, a C++ Array template class library
 * Copyright (C) 1998,2004 Oskar Enoksson (enok@lysator.liu.se)
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */
#ifndef EXPRESSO_TMPLTOOL_HDR
#define EXPRESSO_TMPLTOOL_HDR

namespace TmplTools {
  template<bool B>
  struct TAssert;

  template<>
  struct TAssert<true> {};

  template<>
  struct TAssert<false>;

  template<class T1, class T2>
  struct IsSame {
    enum { check=0 };
  };

  template<class T1>
  struct IsSame<T1,T1> {
    enum { check=1 };
  };

  template<class T0>
  struct UnConst {
    typedef T0 T;
  };

  template<class T0>
  struct UnConst<const T0> {
    typedef T0 T;
  };

  template<class T0>
  struct FlipConst {
    typedef const T0 T;
  };

  template<class T0>
  struct FlipConst<const T0> {
    typedef T0 T;
  };

  template<int N1, int N2, bool B=(N1>N2)>
  struct Max;

  template<int N1, int N2>
  struct Max<N1,N2,true> {
    enum { value = N1 };
  };
  template<int N1, int N2>
  struct Max<N1,N2,false> {
    enum { value = N2 };
  };

  template<class T0>
  struct ArrayInfo {
    typedef T0 RecurseType;
    typedef T0 T;
    enum { len=1 };
    enum { totlen=1 };
    enum { rank=0 };
  };
  
  template<class T0, int M>
  struct ArrayInfo<T0[M]> {
    typedef T0 RecurseType;
    typedef typename ArrayInfo<T0>::T T;
    enum { len=M };
    enum { totlen=len*ArrayInfo<T0>::totlen };
    enum { rank=1+ArrayInfo<T0>::rank };
  };
  
  template<class T0, int N>
  struct ArrayInfoIndex {
    enum { stride=ArrayInfo<T0>::len*
	   ArrayInfoIndex<typename ArrayInfo<T0>::RecurseType,N-1>::stride };
    enum { len=ArrayInfoIndex<typename ArrayInfo<T0>::RecurseType,N-1>::len };
  };
  
  template<class T0>
  struct ArrayInfoIndex<T0,0> {
    enum { stride=1 };
    enum { len=ArrayInfo<T0>::len };
  };

}

#endif
